from __future__ import annotations

import os
import pathlib
from typing import Callable, Iterable, Dict, List, Optional, Tuple, Union

import cv2  # type: ignore
import numpy as np
from PIL import Image  # type: ignore
import tifffile  # type: ignore

from .. import utils

# TODO: TiffReader read slice from inside a page ? (directly in going through the bytes ?)
# Hard and probably not so useful as from what we saw, pages are usually only 2D and therefore not so large


def slice_length(slice_: slice, shape: int) -> int:
    """Compute the number of element in a slice"""
    start, stop, step = slice_.indices(shape)
    return max((stop - start - (step > 0) + (step < 0)) // step + 1, 0)


class MetaVideoReader(type):
    """MetaClass for Video Readers

    Each VideoReader has to define a list of supported extensions.
    The last constructed VideoReader to claim an extension will be used to open the video.
    If no one has claimed an extension the default OpenCVVideoReader is used.
    """

    supported_extensions: List[str] = []
    extension_to_reader: Dict[str, MetaVideoReader] = {}

    def __init__(cls, cls_name: str, bases: tuple, attributes: dict) -> None:
        super().__init__(cls_name, bases, attributes)

        for extension in cls.supported_extensions:
            type(cls).extension_to_reader[extension] = cls


class VideoReader(metaclass=MetaVideoReader):
    """Unified video reader api

    Close to OpenCV API but few key differences:

    * There is always a frame loaded
    * Frame ids goes from 0 to length - 1
    * Frames are loaded in RGB.
    * It support any number of channels and 2D/3D
    * Read method is very different:

        * It retrieves the current frame then grabs the next (The other way around in opencv)
        * It returns therefore a ndarray and a bool rather than a bool and a ndarray
        * The boolean returned indicated if we can continue to read and not if the read operation has failed
    * Easy to check main attributes like:

        * frame_id
        * length
        * channels
        * shape
        * fps if known (-1 otherwise)

    Attributes:
        supported_extensions (List[str]): Static attribute used by `open` method to automatically choose
            which VideoReader to use.
        path (pathlib.Path): Path of the current video
        released (bool): True when release has been called (close and release memory)
        fps (int): Frame rate (-1 if unknown)
        shape (Tuple[int, ...]): Spatial dimensions of frames (Height, Width[, Depth])
        channels (int): Number of channels
        length (int): Number of frames
        frame_id (int): Current frame id

    """

    supported_extensions: List[str] = []

    def __init__(self, path: Union[str, os.PathLike], **kwargs):  # pylint: disable=unused-argument
        """Constructor: Open the video file

        Args:
            path (str | os.PathLike): Path to the video to read
            kwargs: Any additional kwargs are given to the underlying video reader

        """
        self.path = pathlib.Path(path)
        self.released = False
        self.fps = -1
        self.shape: Tuple[int, ...] = (0, 0)
        self.channels = 0
        self.length = 0
        self.frame_id = 0

    def release(self) -> None:
        """Close the file and free memory"""
        self.released = True

    def grab(self) -> bool:
        """Grab the next frame

        Can be faster than self.seek(self.frame_id + 1)

        Returns:
            bool: True if able to grab next frame

        """
        raise NotImplementedError()

    def retrieve(self) -> np.ndarray:
        """Retrieve the current frame

        Returns:
            np.ndarray: The current frame
                Shape: ([D, ]H, W, C)

        """
        raise NotImplementedError()

    def read(self) -> Tuple[np.ndarray, bool]:
        """Consume a frame. Is equivalent to retrieve + grab

        As in this implementation there is always a current frame. It reverses OpenCV implementation
        It first retrieves then grab next frame

        Returns:
            np.ndarray: The current frame - Shape: ([D, ]H, W, C)
            bool: Whether there is a next frame to read

        """
        current_frame = self.retrieve()

        return current_frame, self.grab()

    def seek(self, frame_id: int) -> None:
        """Seek frame_id (will update the current frame)

        Valid frame ids from 0 to length - 1

        Raise:
            EOFError if seeking an invalid frame

        """
        raise NotImplementedError()

    def tell(self) -> int:
        """Returns self.frame_id

        Returns:
            int: current frame_id

        """
        return self.frame_id

    def _check_frame_id(self) -> None:
        """Check that the underlying data is at the right frame id"""

    @staticmethod
    def open(path: Union[str, os.PathLike], **kwargs) -> VideoReader:
        """Open a video file

        Use the extension to know which VideoReader to use

        Args:
            path (str | os.PathLike): File to open
            kwargs: Any additional args for the underlying video reader

        Returns:
            VideoReader

        """
        path = pathlib.Path(path)
        if path.is_dir():
            return MultiFrameReader(path, **kwargs)

        return VideoReader.extension_to_reader.get(path.suffix, OpenCVVideoReader)(path, **kwargs)


class OpenCVVideoReader(VideoReader):
    """Wrapper around opencv VideoCapture

    Default VideoReader when opening a file.

    It only supports 2D images (grayscale or RGB).

    Attributes:
        video (cv2.VideoCapture): VideoCapture from opencv

    """

    def __init__(self, path: Union[str, os.PathLike], **kwargs):
        super().__init__(path, **kwargs)
        self.video = cv2.VideoCapture(str(path), **kwargs)
        assert self.video.grab(), "No frame found in the video"
        self.channels = self.retrieve().shape[-1]

        self.fps = int(self.video.get(cv2.CAP_PROP_FPS))
        self.length = int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))
        self.shape = (int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)), int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)))

    def release(self) -> None:
        super().release()
        self.video.release()

    def grab(self) -> bool:
        grabbed = self.video.grab()
        if grabbed:
            self.frame_id += 1
        else:
            self.seek(self.frame_id)
        return grabbed

    def retrieve(self) -> np.ndarray:
        frame = self.video.retrieve()[1]

        if len(frame.shape) == 2:
            return frame[..., None]

        return np.flip(frame, -1)

    def seek(self, frame_id: int) -> None:
        if not 0 <= frame_id < self.length:
            raise EOFError(f"Seeking outside of video limits: {frame_id} not in [{0}, {self.length}[")
        self.video.set(cv2.CAP_PROP_POS_FRAMES, frame_id + 1)
        self.frame_id = frame_id

    def _check_frame_id(self) -> None:
        assert self.frame_id == self.video.get(cv2.CAP_PROP_POS_FRAMES) - 1


class PILVideoReader(VideoReader):
    """Old PIL video reader. Works well for 2D multi frames Tiff files that are not supported by OpenCV.

    IT only supports 2D videos. TiffVideoReader should have a larger support for TiffFiles

    See `VideoReader` for inherited attributes.

    Attributes:
        video (PIL.Image.Image): PIL image (animated)

    """

    supported_extensions = [".tif", ".tiff"]

    def __init__(self, path: Union[str, os.PathLike], **kwargs):
        super().__init__(path, **kwargs)
        self.video = Image.open(path, **kwargs)
        self.channels = self.retrieve().shape[-1]
        self.shape = self.video.size[::-1]
        self.length = self.video.n_frames  # type: ignore

    def release(self) -> None:
        super().release()
        self.video.close()

    def grab(self) -> bool:
        try:
            self.seek(self.frame_id + 1)
            return True
        except EOFError:
            return False

    def retrieve(self) -> np.ndarray:
        frame = np.array(self.video)

        if len(frame.shape) == 2:
            return frame[..., None]

        return frame

    def seek(self, frame_id: int) -> None:
        self.video.seek(frame_id)
        self.frame_id = frame_id

    def _check_frame_id(self) -> None:
        assert self.frame_id == self.video.tell()


class TiffVideoReader(VideoReader):
    """Tiff video reader with tifffile. Handle 2D and 3D videos with any channels

    Axes are inferred from the tifffile metadata and convert into (T, [D, ]H, W, C) (<=> T[Z]YXC).
    We may not support all formats, or your specific metadata can be wrong/missing. In this case,
    you can also provide the expected axes of the tifffile using an ordered string.

    For example: "TYX" for 2D videos without channel, "TCZYX" for 3D videos with channels
    (ordered by time, channel then stack), "ZTYX" for 3D videos without channels (ordered by stack then time).

    Note:
        With tifffile syntax, we use X for width, Y for height, Z for depth and
        C/S for channels (C and S are not supported together) and T for time.
        Any other letter (I, O, Q, ...) is first interpret as T if it is missing,
        then Z if it is missing, and finally it will yield an error.

    It also supports to read the tiff at a specific resolution level

    See `VideoReader` for inherited attributes.

    Attributes:
        out_axes (str): Axes order of the outputs
        level (int): Resolution level (if any)
            Default: 0 (finest level)
        in_axes (Dict[str, int]): Parsed in axes
        ax_slice (Dict[str, slice]): Optional slices to use on axes.

    """

    supported_extensions = [".tif", ".tiff"]

    out_axes = "TZYXC"  # TZYXC = TDHWC

    def __init__(
        self,
        path: Union[str, os.PathLike],
        level=0,
        axes: Optional[str] = None,
        ax_slice: Optional[Dict[str, slice]] = None,
        **kwargs,
    ):
        """Constructor

        Args:
            axes (Optional[str]): Override the axes found in the tiff metadata.
                Default: None (Parse the metadata)
        """
        super().__init__(path, **kwargs)

        assert self.out_axes[-1] == "C", "Channels have to be the last out axes"

        self.video = tifffile.TiffFile(path, **kwargs).series[0].levels[level]  # pylint: disable=unsubscriptable-object
        self.level = level
        self.ax_slice = ax_slice if ax_slice is not None else {}
        self._page_dim = len(self.video.keyframe.axes)
        self.in_axes = self._extract_axes(axes)

        self._check_slice(self.ax_slice, self.out_axes)

        shape = tuple(
            shape if axis not in self.ax_slice else slice_length(self.ax_slice[axis], shape)
            for shape, axis in zip(self.video._shape_squeezed, self.axes)  # pylint: disable=protected-access
        )

        self.channels = shape[self.in_axes["C"]] if "C" in self.in_axes else 1
        self.shape = tuple(
            shape[self.in_axes[axis]] for axis in self.out_axes if axis in ("X", "Y", "Z") and axis in self.in_axes
        )
        self.length = shape[self.in_axes["T"]]

        self._current = np.zeros((*shape[: self.in_axes["T"]], *shape[self.in_axes["T"] + 1 :]), dtype=self.video.dtype)
        self._load()

    @staticmethod
    def _check_slice(ax_slice: Dict[str, slice], axes: str):
        if "T" in ax_slice:
            raise ValueError("Temporal slicing is not supported for VideoReader. It can be done though Video slicing.")

        for axis, slice_ in ax_slice.items():
            assert axis in axes, f"Unknown axis: {axis} not in {axes}"
            if slice_.step is not None and slice_.step < 1:
                raise ValueError(
                    "Negative step slicing is not supported for VideoReader. It can be done though Video slicing"
                )

    def _extract_axes(self, axes: Optional[str] = None) -> Dict[str, int]:
        if axes:
            # Check that the provided axes are valid
            assert len(set(axes)) == len(axes), "Duplicated axes"
            assert "T" in axes and "X" in axes and "Y" in axes, "Axes should include T, X and Y"
            assert len(axes) <= 5, "The reader can only handle up to 3D + T + C"
            for axis in axes:
                assert axis in self.out_axes, f"Unknown axis: {axis} not in {self.out_axes}"
            assert "T" not in axes[-self._page_dim :], "The temporal axis cannot be in the pages dimensions"
            return {axis: i for i, axis in enumerate(axes)}

        axes_squeezed = self.video._axes_squeezed  # pylint: disable=protected-access
        in_axes = {axis: i for i, axis in enumerate(axes_squeezed)}
        if "S" in in_axes:  # Interpret S as C if possible
            if "C" in in_axes:
                raise ValueError("Unable to parse data with two channel dimensions (C and S)")
            in_axes["C"] = in_axes.pop("S")

        # Check that everything is fine
        assert len(in_axes) == len(axes_squeezed), "Duplicated axes"
        assert 3 <= len(in_axes) <= 5, "The reader expects at least 2D + T and at most 3D + T + C"
        assert "X" in in_axes and "Y" in in_axes, "Missing a spatial dimension X or Y"
        # T is checked afterwards

        for axis in list(in_axes.keys()):
            if axis not in self.out_axes:  # An unexpected key
                if "T" not in in_axes:
                    in_axes["T"] = in_axes.pop(axis)  # Let's assume that it means T first
                elif "Z" not in in_axes:
                    in_axes["Z"] = in_axes.pop(axis)  # Let's assume that it means Z then
                else:
                    raise ValueError(f"Unable to parse the tiff axes: {axis} ({axes_squeezed}) not in {self.out_axes}.")

        if "T" not in in_axes:
            if "Z" not in in_axes:
                raise ValueError(f"Unable to find a temporal axis in {axes_squeezed}")

            in_axes["T"] = in_axes.pop("Z")  # Let's assume that Z is T

        assert in_axes["T"] < len(in_axes) - self._page_dim, "The temporal axis cannot be in the pages dimensions"

        return in_axes

    @property
    def axes(self) -> Iterable[str]:
        return (axis for axis, _ in sorted(self.in_axes.items(), key=lambda k_v: k_v[1]))

    # def _load(self):  # Old load without slices
    #     temporal_axis = self.in_axes["T"]
    #     shape = self.video._shape_squeezed  # pylint: disable=protected-access
    #     before = int(np.prod(shape[: -self._page_dim][:temporal_axis]))
    #     after = int(np.prod(shape[: -self._page_dim][temporal_axis + 1 :]))

    #     # Create a view on current
    #     current = self._current.reshape((before, after, *self.video.keyframe.shape))

    #     for i in range(before):
    #         for j in range(after):
    #             current[i, j] = self.video[(i * self.length + self.frame_id) * after + j].asarray()

    def _load(self):
        temporal_axis = self.in_axes["T"]

        true_shape = self.video._shape_squeezed  # pylint: disable=protected-access
        slices = tuple(slice(None) if axis not in self.ax_slice else self.ax_slice[axis] for axis in self.axes)

        # Compute stride and offset for the out of page dimensions
        strides = np.cumprod((1, *true_shape[-self._page_dim - 1 : 0 : -1]))[::-1]
        offset = np.array([0 if slice_.start is None else slice_.start for slice_ in slices])[: -self._page_dim]
        offset = (offset * strides).sum()
        strides *= np.array([1 if slice_.step is None else slice_.step for slice_ in slices])[: -self._page_dim]

        for index in np.ndindex(*self._current.shape[: -self._page_dim]):
            index_list = list(index)
            index_list.insert(temporal_axis, self.frame_id)

            i = int(offset + np.sum(np.array(index_list) * strides))
            self._current[index] = self.video[i].asarray()[slices[-self._page_dim :]]  # type: ignore

    def release(self) -> None:
        super().release()
        self.video.parent.close()

    def grab(self) -> bool:
        if self.frame_id + 1 >= self.length:
            return False

        self.frame_id += 1
        self._load()
        return True

    def seek(self, frame_id: int) -> None:
        if not 0 <= frame_id < self.length:
            raise EOFError(f"Seeking outside of video limits: {frame_id} not in [{0}, {self.length}[")

        if frame_id != self.frame_id:
            self.frame_id = frame_id
            self._load()

    def retrieve(self) -> np.ndarray:
        # Permute axes into [Z]YXC
        frame = self._current.transpose(
            tuple(
                self.in_axes[axis] - (self.in_axes[axis] > self.in_axes["T"])  # After T => -1 in axis as T is gone
                for axis in self.out_axes
                if axis in self.in_axes and axis != "T"
            )
        )

        if "C" not in self.in_axes:
            frame = frame[..., None]

        # Return a contiguous copy of self._current
        return frame.copy("C")


class FrameTiffLoader:  # pylint: disable=too-few-public-methods
    """Load a single frame stored in a TiffFile with tifffile.

    It handle 2D and 3D videos with any channels. Axes are inferred from the tifffile
    metadata and convert into ([D, ]H, W, C) (<=> [Z]YXC).

    We may not support all formats, or your specific metadata can be wrong/missing. In this case,
    you can also provide the expected axes of the tifffile using an ordered string.

    For example: "YX" for 2D videos without channel, "CZYX" for 3D videos with channels
    (ordered by channel then stack).

    Note:
        With tifffile syntax, we use X for width, Y for height, Z for depth and
        C/S for channels (C and S are not supported together). Any other letter
        (T, I, O, Q, ...) is either intepreted as Z if it is missing, or it will
        yield an error.

    It also supports to read the tiff at a specific resolution level.

    Attributes:
        out_axes (str): Axes order of the outputs
        level (int): Resolution level (if any)
            Default: 0 (finest level)
        axes (Optional[str]): Override the axes found in the tiff metadata.
            Default: None (Parse the metadata)

    """

    out_axes = "ZYXC"  # ZYXC = DHWC

    def __init__(self, level=0, axes: Optional[str] = None, ax_slice: Optional[Dict[str, slice]] = None):
        self.level = level
        self.axes = axes
        self.ax_slice = {} if ax_slice is None else ax_slice

        TiffVideoReader._check_slice(self.ax_slice, self.out_axes)

        assert self.out_axes[-1] == "C", "Channels have to be the last out axis"

        if self.axes:
            # Check that the provided axes are valid

            assert len(set(self.axes)) == len(self.axes), "Duplicated axes"
            assert "X" in self.axes and "Y" in self.axes, "Missing a spatial dimension X or Y"
            assert len(self.axes) <= 4, "The frame loader can only load up to 3D + C images"
            for axis in self.axes:
                assert axis in self.out_axes, f"Unknown axis: {axis} not in {self.out_axes}"

    def _extract_axes(self, serie: tifffile.TiffPageSeries) -> Dict[str, int]:
        if self.axes:
            return {axis: i for i, axis in enumerate(self.axes)}

        axes_squeezed = serie._axes_squeezed  # pylint: disable=protected-access
        axes = {axis: i for i, axis in enumerate(axes_squeezed)}
        if "S" in axes:
            if "C" in axes:
                raise ValueError("Unable to parse data with two channel dimensions (C and S)")
            axes["C"] = axes.pop("S")

        # Check that everything is fine
        assert len(axes) == len(axes_squeezed), "Duplicated axes"
        assert "X" in axes and "Y" in axes, "Missing a spatial dimension X or Y"
        assert len(serie.axes) <= 4, "The frame loader can only load up to 3D + C images"

        for axis in list(axes.keys()):
            if axis not in self.out_axes:  # An unexpected key
                if "Z" not in axes:
                    axes["Z"] = axes.pop(axis)  # Let's assume that it means Z
                else:
                    raise ValueError(f"Unable to parse the tiff axes: {axis} ({axes_squeezed}) not in {self.out_axes}.")

        return axes

    def __call__(self, path: Union[str, os.PathLike]) -> np.ndarray:
        tiff = tifffile.TiffFile(path)
        serie = tiff.series[0].levels[self.level]  # pylint: disable=unsubscriptable-object
        axes = self._extract_axes(serie)

        if not self.ax_slice:
            output = tiff.asarray(series=0, level=self.level, squeeze=True)
            assert output.shape == serie._shape_squeezed, "Unable to read at the expected shape"
        else:
            slices = tuple(
                slice(None) if axis not in self.ax_slice else self.ax_slice[axis]
                for axis in (axis for axis, _ in sorted(axes.items(), key=lambda k_v: k_v[1]))
            )
            output = self._read_slice(serie, slices)

        # Permute axes into [Z]YXC
        output = output.transpose(tuple(axes[axis] for axis in self.out_axes if axis in axes))

        if "C" not in axes:
            output = output[..., None]

        return output

    @staticmethod
    def _read_slice(serie: tifffile.TiffPageSeries, slices: Tuple[slice, ...]) -> np.ndarray:
        """Read the serie using the given slice"""

        page_dim = len(serie.keyframe.axes)

        true_shape = serie._shape_squeezed  # pylint: disable=protected-access
        final_shape = tuple(slice_length(slice_, shape) for slice_, shape in zip(slices, true_shape))

        # Compute stride and offset for the out of page dimensions
        strides = np.cumprod((1, *true_shape[-page_dim - 1 : 0 : -1]))[::-1]
        offset = np.array([0 if slice_.start is None else slice_.start for slice_ in slices])[:-page_dim]
        offset = (offset * strides).sum()
        strides *= np.array([1 if slice_.step is None else slice_.step for slice_ in slices])[:-page_dim]

        output = np.zeros(final_shape, dtype=serie.keyframe.dtype)

        for index in np.ndindex(*final_shape[:-page_dim]):
            i = int(offset + np.sum(np.array(index) * strides))
            output[index] = serie[i].asarray()[slices[-page_dim:]]  # type: ignore

        return output


def pil_loader(path: Union[str, os.PathLike]) -> np.ndarray:
    """Load an image with PIL. Shape: (H, W, C)

    It only supports 2D images
    """
    frame: np.ndarray = np.array(Image.open(path))

    if len(frame.shape) == 2:
        return frame[..., None]

    return frame


class MultiFrameReader(VideoReader):
    """Read video from a list of files inside a folder

    By default, it will find the alphanumerically sorted list of paths
    that shares the most common extension in the folder. The extension may be provided by the user.

    You can provide your own list of paths (absolute paths). The folder path is then ignored.

    Finally, you may also provide your own loading function to load each frame as a numpy array.

    See `VideoReader` for inherited attributes.

    Attributes:
        paths (List[pathlib.Path]): Sorted list of Paths to each frame of the video.
        frame_loader (Callable[[Union[str, os.PathLike]], np.ndarray]): Loads frame from their associated files.

    """

    def __init__(
        self,
        path: Union[str, os.PathLike],
        paths: Optional[List[Union[str, os.PathLike]]] = None,
        extension: Optional[str] = None,
        frame_loader: Optional[Callable[[Union[str, os.PathLike]], np.ndarray]] = None,
        **kwargs,
    ):
        super().__init__(path, **kwargs)

        if paths:
            self.paths = [pathlib.Path(path) for path in paths]
            assert len({path.suffix for path in self.paths}) == 1, "Found several extensions in paths"
        else:
            files = list(file for file in self.path.iterdir() if file.is_file())
            if extension is None:
                extensions: Dict[str, int] = {}
                for file in files:
                    extensions[file.suffix] = extensions.get(file.suffix, 0) + 1

                extension = max(extensions.items(), key=lambda k_v: k_v[1])[0]

            self.paths = utils.sorted_alphanumeric(file for file in files if file.suffix == extension)

            assert self.paths, "No frame found in the video"

        if frame_loader is None:
            if self.paths[0].suffix in (".tif", ".tiff"):
                frame_loader = FrameTiffLoader(**kwargs)
            else:
                frame_loader = pil_loader
        self.frame_loader = frame_loader

        self.length = len(self.paths)

        first_frame = self.retrieve()
        self.shape = first_frame.shape[:-1]
        self.channels = first_frame.shape[-1]

    def grab(self) -> bool:
        if self.frame_id + 1 >= self.length:
            return False

        self.frame_id += 1
        return True

    def seek(self, frame_id: int) -> None:
        if not 0 <= frame_id < self.length:
            raise EOFError(f"Seeking outside of video limits: {frame_id} not in [{0}, {self.length}[")

        self.frame_id = frame_id

    def retrieve(self) -> np.ndarray:
        # TODO: Do the read in grab/seek and store in a _current attribute to have a fast retrieve ?
        return self.frame_loader(self.paths[self.frame_id])



import dataclasses
from typing import Optional, Tuple, Union
import warnings

import numpy as np
import torch

import byotrack
from .base import AssociationMethod, FrameByFrameLinker, FrameByFrameLinkerParameters


@dataclasses.dataclass
class NearestNeighborParameters(FrameByFrameLinkerParameters):
    """Parameters of NearestNeighborLinker

    Note:
        The merging and splitting features is still experimental.

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
            Default: 5 pixels
        n_valid (int): Number associated detections required to validate the track after its creation.
            Default: 3
        n_gap (int): Number of consecutive frames without association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
        anisotropy (Tuple[float, float, float]): Anisotropy of images (Ratio of the pixel sizes
            for each axis, depth first). This will be used to scale distances.
            Default: (1., 1., 1.)
        ema (float): Optional exponential moving average to reduce detection noise. Detection positions are smoothed
            using this EMA. Should be smaller than 1. It use: x_{t+1} = ema x_{t} + (1 - ema) det(t)
            As motion is not modeled, EMA may introduce lag that will hinder tracking. It is more effective with
            optical flow to compensate motions, in this case, a typical value is 0.5, to average the previous position
            with the current measured one. For more advanced modelisation, see `KalmanLinker`.
            Default: 0.0 (No EMA)
        fill_gap (bool): Fill the gap of missed detections using a forward optical flow
            propagation (Only when optical flow is provided). We advise to rather use a
            ForwardBackward interpolation using the same optical flow: it will produce
            smoother interpolations.
            Default: False
        split_factor (float): Allow splitting of tracks, using a second association step.
            The association threshold in this case is `split_factor * association_threshold`.
            Default: 0.0 (No splits)
        merge_factor (float): Allow merging of tracks, using a second association step.
            The association threshold in this case is `merge_factor * association_threshold`.
            Default: 0.0 (No merges)

    """

    def __init__(
        self,
        association_threshold: float = 5.0,
        *,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        ema=0.0,
        fill_gap=False,
        split_factor: float = 0.0,
        merge_factor: float = 0.0,
    ):
        super().__init__(  # pylint: disable=duplicate-code
            association_threshold=association_threshold,
            n_valid=n_valid,
            n_gap=n_gap,
            association_method=association_method,
            anisotropy=anisotropy,
            split_factor=split_factor,
            merge_factor=merge_factor,
        )
        self.ema = ema
        self.fill_gap = fill_gap

    ema: float = 0.0
    fill_gap: bool = False


class NearestNeighborLinker(FrameByFrameLinker):
    """Frame by frame linker by associating with the closest detections

    Motion is not modeled, but if an optical flow method is provided, it
    will be used to compensate motion online. Matching is done from a simple Euclidean
    distance. This can be easily changed by inheriting this class and overriding the `cost` method.

    See `FrameByFrameLinker` for the other attributes.

    Attributes:
        specs (NearestNeighborParameters): Parameters specifications of the algorithm.
            See `NearestNeighborParameters`.
        active_positions (Optional[torch.Tensor]): The positions of actives tracks,
            if undetected it is estimated by optical flow propagation.
            Shape: (N, D), dtype: float32

    """

    progress_bar_description = "Nearest Neighbor linking"

    def __init__(
        self,
        specs: NearestNeighborParameters,
        optflow: Optional[byotrack.OpticalFlow] = None,
        features_extractor: Optional[byotrack.FeaturesExtractor] = None,
        save_all=False,
    ) -> None:
        super().__init__(specs, optflow, features_extractor, save_all)
        self.specs: NearestNeighborParameters
        self.active_positions: Optional[torch.Tensor] = None

        if self.specs.fill_gap and not self.optflow:
            warnings.warn("Optical flow has not been provided. Gap cannot be filled")

    def reset(self) -> None:
        super().reset()
        self.active_positions = None

    def motion_model(self) -> None:
        if self.optflow and self.optflow.flow_map is not None:
            if self.active_positions is not None:
                self.active_positions = torch.tensor(
                    self.optflow.optflow.transform(self.optflow.flow_map, self.active_positions.numpy())
                )

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        if self.active_positions is None:
            self.active_positions = torch.empty((0, detections.position.shape[1]))

        anisotropy = torch.tensor(self.specs.anisotropy)[: -detections.dim]

        return (
            torch.cdist(self.active_positions * anisotropy, detections.position * anisotropy),
            self.specs.association_threshold,
        )

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, active_mask: torch.Tensor):
        if self.active_positions is None:
            self.active_positions = torch.empty((0, detections.position.shape[1]))

        # Update tracks positions with detections
        # Optionally using an EMA to reduce detections noise
        self.active_positions[self._links[:, 0]] -= (1.0 - self.specs.ema) * (
            self.active_positions[self._links[:, 0]] - detections.position[self._links[:, 1]]
        )

        # Merge still active positions and new ones
        self.active_positions = torch.cat(
            (self.active_positions[active_mask], detections.position[self._unmatched_detections])
        )

        self.all_positions.append(self.active_positions.clone())

        if not self.specs.fill_gap:  # Erase optical flow predictions from the stored positions
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] == -1:  # Not truly detected
                    self.all_positions[-1][i, :] = torch.nan



from abc import abstractmethod
import dataclasses
from typing import Dict, List, Optional, Tuple, Union
import warnings

import enum

import numpy as np
import pylapy
import torch

import byotrack

from .greedy_lap import greedy_assignment_solver


class AssociationMethod(enum.Enum):
    """Association methods (Greedy or Jonker-Volgenant)

    * GREEDY
        Select the best match between tracks and detections iteratively until
        no match can be selected below the cost limit eta. It is usually not optimal for tracking
        but it is much faster.
    * OPT_HARD | SPARSE_OPT_HARD
        Solve the linear association problem (see `pylapy`).
        Hard threshold the association matrix with the cost limit eta.
        Use the sparse version to increase speed with numerous particles.
    * OPT_SMOOTH | SPARSE_OPT_SMOOTH
        Solve a cost_limit extended association problem (see `pylapy`)
        It relaxes the linear association problem, allowing to not link a node
        for the cost limit eta.
        Use the sparse version to increase speed with numerous particles.
    """

    GREEDY = "greedy"
    OPT_HARD = "opt_hard"
    OPT_SMOOTH = "opt_smooth"
    SPARSE_OPT_HARD = "sparse_opt_hard"
    SPARSE_OPT_SMOOTH = "sparse_opt_smooth"

    def solve(self, cost: torch.Tensor, eta: float = np.inf) -> torch.Tensor:
        """Solve tracks-to-detections association

        Args:
            cost (torch.Tensor): Cost matrix
                    Shape: (N, M), dtype: float
            eta (float): Cost limit
                Default: inf (No thresholding)

        Returns:
            torch.Tensor: Links (i, j)
                Shape: (L, 2), dtype: int32
        """
        if self == AssociationMethod.SPARSE_OPT_HARD:
            return torch.tensor(pylapy.LapSolver().sparse_solve(cost.numpy(), eta, hard=True).astype(np.int32))

        if self == AssociationMethod.SPARSE_OPT_SMOOTH:
            return torch.tensor(pylapy.LapSolver().sparse_solve(cost.numpy(), eta, hard=False).astype(np.int32))

        if self == AssociationMethod.OPT_HARD:
            cost[cost > eta] = np.inf
            return torch.tensor(pylapy.LapSolver().solve(cost.numpy()).astype(np.int32))

        if self == AssociationMethod.OPT_SMOOTH:
            return torch.tensor(pylapy.LapSolver().solve(cost.numpy(), eta).astype(np.int32))

        return torch.tensor(greedy_assignment_solver(cost.numpy(), eta).astype(np.int32))


class TrackHandler:  # pylint: disable=too-many-instance-attributes
    """Handle a track during the tracking procedure

    It accumulates the track data at each new association and store the optional motion model data.

    A TrackHandler is created for each unlinked detections in the linking process and then updated with
    the following associated detections. At the beginining, the track is considered HYPOTHETICAL.
    For a track to be considered valid, it requires n_valid consecutive associated detections after the track creation
    (state: HYPOTHETICAL => VALID). It a miss detection occurs during this time interval, then the track is deleted
    and considered invalid (state: HYPOTHETICAL => INVALID).

    Once confirmed, a VALID track is resilient to miss detections, waiting n_gap frames before ending the track
    (VALID => FINISHED).

    Attributes:
        n_valid (int): Number of frames with a correct association required to validate the track at its creation.
        n_gap (int): Number of frames with no association before the track termination.
        start (int): Starting frame of the track
        identifier (int): Identifier of the track handler (and of the track)
        track_state (TrackState): Current state of the handler
        last_association (int): Number of frames since the last association
        detection_ids (List[int]): Identifiers of the associated detection (-1 if None)
        track_ids (List[int]): Index of the track at each frame in the `linker.active_tracks` list.
            It allows the linker to store data as tensor and be able to rebuild tracks at the end.
        merge_id (int): Identifier to an optional merged track handler (See `Tracks.merge_id`)
        parent_id (int): Identifier to an optional parent track handler (See `Tracks.parent_id`)
        is_split (bool): Just to know if the track splits

    """

    class TrackState(enum.IntEnum):
        """TrackState of a TrackHandler

        * HYPOTHETICAL
            Initial state before validation of the track.
        * VALID
            The track has been validated and is still active
        * FINISHED
            The track is valid and finished
        * INVALID
            The track is not valid and deleted

        """

        HYPOTHETICAL = 0
        VALID = 1
        FINISHED = 2
        INVALID = 3

    def __init__(self, n_valid: int, n_gap: int, start: int, identifier: int) -> None:
        self.n_valid = n_valid
        self.n_gap = n_gap
        self.start = start
        self.identifier = identifier
        self.track_state = TrackHandler.TrackState.HYPOTHETICAL
        self.last_association = 0
        self.detection_ids: List[int] = []
        self.track_ids: List[int] = []
        self.merge_id = -1
        self.parent_id = -1
        self.is_split = False

    def __len__(self) -> int:
        if self.merge_id != -1 or self.is_split:
            return len(self.detection_ids)  # Last points counts

        return len(self.detection_ids) - self.last_association

    def is_active(self) -> bool:
        return self.track_state < 2

    def update(self, frame_id: int, detection_id: int) -> None:
        """Update track handler. It stores the detection_id and update the track state.

        It should be called for each time frame and each active track.

        Args:
            frame_id (int): The current frame. This is given for safety checks
                to ensure that the Linker and TrackHandler agree.
            detection_id (int): Detection id in the Detections object.
                -1 if not associated to a particular detection.

        """
        assert self.is_active()
        assert len(self.track_ids) == len(
            self.detection_ids
        ), "The linker should call `update` then `register_track_id` at each linking step"
        assert (
            self.start + len(self.detection_ids) == frame_id
        ), "The linker should update each active track on each time frame."

        self.detection_ids.append(detection_id)

        if detection_id == -1:  # Not associated
            self.last_association += 1

            if self.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                self.track_state = TrackHandler.TrackState.INVALID

            if self.last_association > self.n_gap:
                self.track_state = TrackHandler.TrackState.FINISHED
        else:
            self.last_association = 0

            if self.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                if len(self.detection_ids) >= self.n_valid:
                    self.track_state = TrackHandler.TrackState.VALID

    def register_track_id(self, track_id: int) -> None:
        """For still active tracks, it registers the track id after the update step.

        Args:
            track_id (int): The index of the track in `linker.active_tracks` at this time frame.
        """
        self.track_ids.append(track_id)


class OnlineFlowExtractor:
    """Extract optical flow maps online from a video"""

    def __init__(self, optflow: byotrack.OpticalFlow) -> None:
        self.optflow = optflow
        self.flow_map: Optional[np.ndarray] = None
        self.src: Optional[np.ndarray] = None

    def reset(self) -> None:
        """Reset the flow extractor"""
        self.flow_map = None
        self.src = None

    def update(self, frame: np.ndarray) -> None:
        """Extract the flow for a new given frame

        It will compute and register the flow map between the last given frame
        and the current frame.

        Args:
            frame (np.ndarray): Current frame of the video

        """
        dst = self.optflow.preprocess(frame)
        if self.src is None:
            self.src = dst

        self.flow_map = self.optflow.compute(self.src, dst)
        self.src = dst


@dataclasses.dataclass
class FrameByFrameLinkerParameters:  # pylint: disable=too-many-instance-attributes
    """Parameters of the abstract FrameByFrameLinker

    Note:
        The merging and splitting features is still experimental.

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
            Default: 5 pixels
        n_valid (int): Number associated detections required to validate the track after its creation.
            Default: 3
        n_gap (int): Number of consecutive frames without association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, OPT_HARD, OPT_SMOOTH)
            Default: OPT_SMOOTH
        anisotropy (Tuple[float, float, float]): Anisotropy of images (Ratio of the pixel sizes
            for each axis, depth first). This will be used to scale distances.
            Default: (1., 1., 1.)
        split_factor (float): Allow splitting of tracks, using a second association step.
            The association threshold in this case is `split_factor * association_threshold`.
            Default: 0.0 (No splits)
        merge_factor (float): Allow merging of tracks, using a second association step.
            The association threshold in this case is `merge_factor * association_threshold`.
            Default: 0.0 (No merges)

    """

    def __init__(
        self,
        association_threshold: float = 5.0,
        *,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        split_factor: float = 0.0,
        merge_factor: float = 0.0,
    ):
        self.association_threshold = association_threshold
        self.n_valid = n_valid
        self.n_gap = n_gap
        self.association_method = (
            association_method
            if isinstance(association_method, AssociationMethod)
            else AssociationMethod[association_method.upper()]
        )
        self.anisotropy = anisotropy
        self.split_factor = split_factor
        self.merge_factor = merge_factor

        if merge_factor > 1.0 or split_factor > 1.0:
            warnings.warn("Merge or split factors should be lower than 1")

    association_threshold: float = 5.0
    n_valid: int = 3
    n_gap: int = 3
    association_method: AssociationMethod = AssociationMethod.OPT_SMOOTH
    anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0)
    split_factor: float = 0.0
    merge_factor: float = 0.0


class FrameByFrameLinker(byotrack.OnlineLinker):  # pylint: disable=too-many-instance-attributes
    """Links detections online using frame-by-frame association

    Abstract class for frame-by-frame linker. It decomposes the update step in 6 parts:

    1. Optional optical flow computations (Handled by this class with the `optflow` given)
    2. Motion modeling to predict track positions (`motion_model`)
    3. Features extraction (handled by this class with the `features_extractor` given)
    4. Track-to-detection cost computation (`cost`)
    5. Solving the linear association problem (handled in `associate`)
    6. Post matching update to handle tracks (`post_association`)

    The association relies on the AssociationMethod enum and tracks handling is done with
    TrackHandler.

    It follows the tracks handling strategy describe in KOFT [9].

    Attributes:
        specs (FrameByFrameLinkerParameters): Parameters specifications of the algorithm.
            See `FrameByFrameLinkerParameters`.
        optflow (Optional[OnlineFlowExtractor]): Optional wrapper around the given optional OpticalFlow that
            will extract flow maps of the video online. (The underlying OpticalFlow object is accessible in
            self.optflow.optflow)
            Default: None
        features_extractor (Optional[FeaturesExtractor]): Optional features extractor that will extract
            features for the detections, which could be useful for tracking.
            Default: None
        save_all (bool): Save metadata useless for the final building of tracks
            but that could be useful for analysis. For instance, it will keep invalid tracks.
            Or the computed features inside the Detections objects.
        frame_id (int): Current frame id of the linking process
        inactive_tracks (List[TrackHandler]): Terminated tracks
        active_tracks (List[TrackHandler]): Current track handlers
        all_positions (List[torch.Tensor]): Positions of the active tracks at each seen frames.
            Using the valid track handlers `track_ids`, it allows the reconstruction of tracks.
        split_links (torch.Tensor): Current split_links
            shape: (L', 2), dtype: int32
        merge_links (torch.Tensor): Current merge_links
            shape: (L'', 2), dtype: int32

    """

    def __init__(
        self,
        specs: FrameByFrameLinkerParameters,
        optflow: Optional[byotrack.OpticalFlow] = None,
        features_extractor: Optional[byotrack.FeaturesExtractor] = None,
        save_all=False,
    ) -> None:
        super().__init__()
        self.specs = specs
        self.optflow = OnlineFlowExtractor(optflow) if optflow else None
        self.features_extractor = features_extractor
        self.save_all = save_all
        self.frame_id = -1
        self.inactive_tracks: List[TrackHandler] = []
        self.active_tracks: List[TrackHandler] = []
        self.all_positions: List[torch.Tensor] = []
        self.active_mass = torch.zeros((0,), dtype=torch.float32)

        # Instantaneous useful quantities
        self._links = torch.zeros((0, 2), dtype=torch.int32)
        self._split_links = torch.zeros((0, 2), dtype=torch.int32)
        self._merge_links = torch.zeros((0, 2), dtype=torch.int32)
        self._unmatched_detections = torch.full((0,), True)
        self._next_identifier = 0

    def reset(self) -> None:
        super().reset()
        if self.optflow:
            self.optflow.reset()
        self.frame_id = -1
        self.inactive_tracks = []
        self.active_tracks = []
        self.all_positions = []
        self.active_mass = torch.zeros((0,), dtype=torch.float32)
        self._links = torch.zeros((0, 2), dtype=torch.int32)
        self._split_links = torch.zeros((0, 2), dtype=torch.int32)
        self._merge_links = torch.zeros((0, 2), dtype=torch.int32)
        self._unmatched_detections = torch.full((0,), True)
        self._next_identifier = 0

    def collect(self) -> List[byotrack.Track]:
        tracks = []
        for handler in self.inactive_tracks + self.active_tracks:
            if handler.track_state in (handler.TrackState.INVALID, handler.TrackState.HYPOTHETICAL):
                continue  # Ignore non-valid tracks

            points = torch.cat(
                [
                    positions[track_id : track_id + 1]
                    for track_id, positions in zip(
                        handler.track_ids[: len(handler)], self.all_positions[handler.start :]
                    )
                ]
            )

            tracks.append(
                byotrack.Track(
                    handler.start,
                    points,
                    handler.identifier,
                    torch.tensor(handler.detection_ids[: len(handler)], dtype=torch.int32),
                    parent_id=handler.parent_id,
                    merge_id=handler.merge_id,
                )
            )
        return tracks

    @abstractmethod
    def motion_model(self) -> None:
        """Optional modelisation of motion for tracks

        It can be used to update some internal state of the tracker after the optical flow computation
        and before the distance computation.
        """

    @abstractmethod
    def cost(self, frame: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        """Compute the association cost between active tracks and detections

        It also returns the threshold to use (Depending on the dist you use, association_threshold
        could be related to a more meaning full quantity than the cost itself).
        For instance, when using a squared Euclidean distance, the association threshold could be
        express as the distance in pixel, and this function could square it.
        For likelihood association, you could provide the association threshold as a probability
        and use -log(threshold) as the true threshold. (See `KalmanLinker` and `NearestNeighborLinker`)

        Args:
            frame (np.ndarray): The current frame of the video
                Shape: (H, W, C), dtype: float
            detections (byotrack.Detections): Detections for the given frame

        Returns:
            torch.Tensor: The cost matrix between active tracks and detections
                Shape: (n_tracks, n_dets), dtype: float32
            float: The association threshold to use.
                It can be different than `self.association_threshold` depeding on the dist build here

        """

    @abstractmethod
    def post_association(self, frame: np.ndarray, detections: byotrack.Detections, active_mask: torch.Tensor):
        """Update the internal state of the tracker after `update_active_tracks`

        It should update any internal model/data. It is also responsible to register the position of each active
        track in `all_positions` for the current time frame.

        Args:
            frame (np.ndarray): The current frame of the video
                Shape: (H, W, C), dtype: float
            detections (byotrack.Detections): Detections for the given frame
            active_mask (torch.Tensor): Boolean tensor indicating True for still active tracks
                Shape: (N_tracks), dtype: bool

        """

    def update_active_tracks(  # pylint: disable=too-many-branches,too-many-statements,too-many-locals
        self, detections: byotrack.Detections
    ) -> torch.Tensor:
        """Updates tracks handler and creates new ones for extra detections

        Tracks that are terminated are stored inside `inactive_tracks` and dropped from `active_tracks`.
        It is called by update before `post_association`.

        It also handles merges and splits. In the case of some specific merges, it may change a few links.
        The updated links are returned, with a still_active mask for tracks and a new_track mask
        for detections.

        Args:
            detections (byotrack.Detections): Detections for the given frame

        Returns:
            torch.Tensor: Boolean tensor indicating True for still active tracks
                Shape: (N_tracks), dtype: bool

        """
        if self._split_links.shape[0] + self._merge_links.shape[0] == 0:  # Fall back to old simpler version
            active_mask = self._update_active_tracks()
            self._handle_extra_detections(detections)
            return active_mask

        # Ugly to handle merge and splits smoothly... If you do not care about that, read only the old simpler version
        # Or even use offline split and merge strategies.

        # Create new tracks from unmatched measures
        new_tracks: List[TrackHandler] = []
        for i in torch.arange(len(detections))[self._unmatched_detections].tolist():
            track = TrackHandler(
                self.specs.n_valid,
                self.specs.n_gap,
                self.frame_id,
                self._next_identifier,
            )
            self._next_identifier += 1
            track.update(self.frame_id, i)
            new_tracks.append(track)

        # Lots of useful identifiers mapping
        det_to_new_track = torch.full((len(detections),), -1, dtype=torch.int32)
        det_to_new_track[torch.arange(len(detections))[self._unmatched_detections]] = torch.arange(
            len(new_tracks), dtype=torch.int32
        )

        track_to_det = torch.full((len(self.active_tracks),), -1, dtype=torch.int32)
        track_to_det[self._links[:, 0]] = self._links[:, 1]
        det_to_track = torch.full((len(detections),), -1, dtype=torch.int32)
        det_to_track[self._links[:, 1]] = self._links[:, 0]

        track_to_det_split = torch.full((len(self.active_tracks),), -1, dtype=torch.int32)
        track_to_det_split[self._split_links[:, 0]] = self._split_links[:, 1]

        track_to_det_merge = torch.full((len(self.active_tracks),), -1, dtype=torch.int32)
        track_to_det_merge[self._merge_links[:, 0]] = self._merge_links[:, 1]
        det_to_track_merge = torch.full((len(detections),), -1, dtype=torch.int32)
        det_to_track_merge[self._merge_links[:, 1]] = self._merge_links[:, 0]

        # Update active tracks (for merges and splits, they are still active but replaced by a new handler)
        active_mask = torch.full((len(self.active_tracks),), False)
        still_active: List[TrackHandler] = []
        merges_to_ref: List[TrackHandler] = []
        det_to_merge_id: Dict[int, int] = {}

        for i, track in enumerate(self.active_tracks):
            if track_to_det_split[i] != -1:  # The track splits
                # For the corner case of a hypothetical track, it is now valid (afterall we found 2 detections for it)
                track.track_state = TrackHandler.TrackState.FINISHED
                track.is_split = True
                self.inactive_tracks.append(track)

                # Replace by a new VALID handler
                other = TrackHandler(self.specs.n_valid, self.specs.n_gap, self.frame_id, self._next_identifier)
                self._next_identifier += 1
                other.track_state = TrackHandler.TrackState.VALID

                # Set parent for both new tracks
                other.parent_id = track.identifier
                new_tracks[det_to_new_track[track_to_det_split[i]]].parent_id = track.identifier

                # Swap other with track (the splitted one is finished)
                track = other

            if track_to_det[i] == -1 and track_to_det_merge[i] != -1:  # The track is the second branch of a merge
                assert det_to_track[track_to_det_merge[i]] != -1
                # Get the first branch
                other = self.active_tracks[det_to_track[track_to_det_merge[i]]]

                # If hypothetical, nothing to do, we do not merge with hypothetical tracks
                if track.track_state != TrackHandler.TrackState.HYPOTHETICAL:
                    if other.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                        # If other is hypothetical, we do not merge but we swap the link
                        # This invalidate some mappings for both 'track' and 'other', but it is fine.
                        # If 2nd branch is first to be executed this way, the 1st branch will not be executed
                        track_to_det[det_to_track[track_to_det_merge[i]]] = -1
                        track_to_det[i] = track_to_det_merge[i]
                    else:  # Merge (as this is the second track, the track is just dropped)
                        track.merge_id = int(track_to_det_merge[i])  # This is not merge id yet
                        merges_to_ref.append(track)  # It will be updated once all tracks has been processed

                        track.track_state = TrackHandler.TrackState.FINISHED
                        self.inactive_tracks.append(track)
                        continue  # It does itw own specific update, the track cannot be updated any more.

            if track_to_det[i] != -1 and det_to_track_merge[track_to_det[i]] != -1:  # First branch of a merge
                other = self.active_tracks[det_to_track_merge[track_to_det[i]]]

                # If other is hypothetical, nothing to do, we do not merge with hypothetical tracks
                if other.track_state != TrackHandler.TrackState.HYPOTHETICAL:
                    if track.track_state == TrackHandler.TrackState.HYPOTHETICAL:
                        # If track is hypothetical, we do not merge but we swap the link
                        # If 1st branch is first to be executed this way, the 2nd branch will not be executed
                        track_to_det[det_to_track_merge[track_to_det[i]]] = track_to_det[i]
                        track_to_det[i] = -1
                    else:  # Merge: Create a new track and stop the former one
                        # Replace by a new VALID handler
                        other = TrackHandler(self.specs.n_valid, self.specs.n_gap, self.frame_id, self._next_identifier)
                        self._next_identifier += 1
                        other.track_state = TrackHandler.TrackState.VALID

                        track.merge_id = other.identifier
                        det_to_merge_id[int(track_to_det[i])] = other.identifier

                        # Terminate old track
                        track.track_state = TrackHandler.TrackState.FINISHED
                        self.inactive_tracks.append(track)

                        # Swap other with track (the merged one is finished)
                        track = other  # Will be updated and kept in active

            # Update track (classical link/no link, or a newly created one at a split/merge event)
            track.update(self.frame_id, int(track_to_det[i].item()))

            # Check if track is still active
            if track.is_active():
                still_active.append(track)
                active_mask[i] = True
            elif track.track_state == TrackHandler.TrackState.FINISHED or self.save_all:
                self.inactive_tracks.append(track)

            if track.track_state == TrackHandler.TrackState.INVALID and track.parent_id != -1:
                self._undo_split(track, still_active)

        # Relabel merges
        for track in merges_to_ref:
            track.merge_id = det_to_merge_id[track.merge_id]

        # Relabel links
        self._links[:, 0] = torch.arange(len(track_to_det))[track_to_det != -1]
        self._links[:, 1] = track_to_det[track_to_det != -1]

        self.active_tracks = still_active + new_tracks

        return active_mask

    def _update_active_tracks(self) -> torch.Tensor:
        """Calls `update` for active tracks and return a boolean mask that indicates which track is still active

        Tracks that are terminated are stored inside `inactive_tracks` and dropped from `active_tracks`.

        Returns:
            torch.Tensor: Boolean tensor indicating True for still active tracks

        """
        i_to_j = torch.full((len(self.active_tracks),), -1, dtype=torch.int32)
        i_to_j[self._links[:, 0]] = self._links[:, 1]
        active_mask = torch.full((len(self.active_tracks),), False)
        still_active = []
        for i, track in enumerate(self.active_tracks):
            track.update(self.frame_id, int(i_to_j[i].item()))

            # Check if track is still active
            if track.is_active():
                still_active.append(track)
                active_mask[i] = True
            elif track.track_state == TrackHandler.TrackState.FINISHED or self.save_all:
                self.inactive_tracks.append(track)

            if track.track_state == TrackHandler.TrackState.INVALID and track.parent_id != -1:
                self._undo_split(track, still_active)

        self.active_tracks = still_active

        return active_mask

    def _undo_split(self, track: TrackHandler, still_active: List[TrackHandler]):
        """Undo a split if one of the child track is invalidated.

        It finds the finished parent track handler and the valid other child handler and
        concatenates them back together in a single valid track handler.

        This is quite slow, but this should not occur often.
        """
        # This is slow, this should not occurs often. To be really faster it requires other data structures
        # To be changed if it is necessary

        # Find parent and sibling
        parent = [other for other in self.inactive_tracks if other.identifier == track.parent_id][0]
        child = [other for other in self.active_tracks + self.inactive_tracks if other.parent_id == track.parent_id][0]

        # Remove parent from inactive tracks
        self.inactive_tracks.remove(parent)

        if child.is_active():
            # By construction, the valid child is in still_active (because added before the hypothetical child)
            # So no need to check self.active_tracks
            # if child not in still_active:
            #     still_active = self.active_tracks  # In the loop, we haven't found
            child_index = still_active.index(child)
        else:
            child_index = self.inactive_tracks.index(child)

        concatenated_handler = TrackHandler(self.specs.n_valid, self.specs.n_gap, parent.start, child.identifier)
        concatenated_handler.track_state = child.track_state
        concatenated_handler.last_association = child.last_association
        concatenated_handler.parent_id = parent.parent_id
        concatenated_handler.merge_id = child.merge_id
        concatenated_handler.track_ids = parent.track_ids + child.track_ids
        concatenated_handler.detection_ids = parent.detection_ids + child.detection_ids
        concatenated_handler.is_split = child.is_split

        # Replace child by concatenated
        if concatenated_handler.is_active():
            still_active[child_index] = concatenated_handler
        else:
            self.inactive_tracks[child_index] = concatenated_handler

    def _handle_extra_detections(self, detections: byotrack.Detections):
        """Handle extra detections by creating new track handlers

        Args:
            detections (byotrack.Detections): Detections for the given frame

        """
        # Create a new active track for each unmatched measure
        for i in torch.arange(len(detections))[self._unmatched_detections].tolist():
            handler = TrackHandler(self.specs.n_valid, self.specs.n_gap, self.frame_id, self._next_identifier)
            self._next_identifier += 1
            handler.update(self.frame_id, i)
            self.active_tracks.append(handler)

    def update_detections(self, detections: byotrack.Detections) -> byotrack.Detections:
        """Optional modification of the currrent detections based on the current state

        This is called by `update` after the motion modeling but before the cost/association.

        By default, it does not change anything.

        Args:
            detections (byotrack.Detections): Detections at the current frame

        Returns:
            byotrack.Detections: The (optionally modified) detections to use at this current frame

        """
        return detections

    def associate(self, frame: np.ndarray, detections: byotrack.Detections) -> torch.Tensor:
        """Produces links between the current tracks and detections

        Optionnally it handles merges and splits by assocating a second time.

        Args:
            frame (np.ndarray): Current frame
            detections (byotrack.Detections): Current detections

        Returns:
            torch.Tensor: Links (i, j)
                Shape: (L, 2), dtype: int32
        """

        cost, threshold = self.cost(frame, detections)
        self._links = self.specs.association_method.solve(cost, threshold)

        self._unmatched_detections = torch.full((len(detections),), True)
        self._unmatched_detections[self._links[:, 1]] = False

        if self.specs.merge_factor == 0 and self.specs.split_factor == 0:
            return self._links  # No merge or splits

        unmatched_tracks = torch.full((len(self.active_tracks),), True)
        unmatched_tracks[self._links[:, 0]] = False
        valid_tracks = torch.tensor(
            [(track.track_state == TrackHandler.TrackState.VALID) for track in self.active_tracks], dtype=torch.bool
        )

        if self.specs.merge_factor > 0:
            # We simply do a 2nd association between unassociated VALID tracks with associated detections
            tracks_mask = unmatched_tracks & valid_tracks

            # TODO: Mass factor
            # self.active_mass[tracks_mask]

            self._merge_links = self.specs.association_method.solve(
                cost[tracks_mask][:, ~self._unmatched_detections], threshold * self.specs.merge_factor
            )

            # Relabel
            self._merge_links[:, 0] = torch.arange(len(self.active_tracks))[tracks_mask][self._merge_links[:, 0]]
            self._merge_links[:, 1] = torch.arange(len(detections))[~self._unmatched_detections][
                self._merge_links[:, 1]
            ]

        if self.specs.split_factor > 0:
            # We simply do a 2nd association between associated tracks with unassociated detections

            # Split mass factor
            # We increase the distance if the split is not evenly weighted and if it does not sum at the previous mass
            track_mass = self.active_mass[self._links[:, 0]]
            associated_mass = detections.mass[self._links[:, 1]]
            non_associated_mass = detections.mass[self._unmatched_detections]

            even_factor = torch.maximum(associated_mass[:, None], non_associated_mass[None, :]) / torch.minimum(
                associated_mass[:, None], non_associated_mass[None, :]
            )
            sum_ = associated_mass[:, None] + non_associated_mass[None, :]
            mass_factor = torch.maximum(track_mass[:, None], sum_) / torch.minimum(track_mass[:, None], sum_)

            self._split_links = self.specs.association_method.solve(
                cost[~unmatched_tracks][:, self._unmatched_detections] * even_factor * mass_factor,
                threshold * self.specs.split_factor,
            )

            # Relabel
            self._split_links[:, 0] = torch.arange(len(self.active_tracks))[~unmatched_tracks][self._split_links[:, 0]]
            self._split_links[:, 1] = torch.arange(len(detections))[self._unmatched_detections][self._split_links[:, 1]]

        return self._links

    def update(self, frame: np.ndarray, detections: byotrack.Detections) -> None:
        self.frame_id += 1

        # Compute the flow map if optflow given
        if self.optflow is not None:
            self.optflow.update(frame)

        self.motion_model()
        detections = self.update_detections(detections)

        # Compute features if the extractor is given and register inside the detections
        # Do not recompute the features if some are already registered
        remove_feats = False
        if self.features_extractor is not None:
            if "features" in detections.data:
                warnings.warn("Some features are already computed. They will be used.")
            else:
                remove_feats = True
                self.features_extractor.register(frame, detections)

        self.associate(frame, detections)
        active_mask = self.update_active_tracks(detections)
        self.post_association(frame, detections, active_mask)

        # Handle mass with a fixed ema and concatenate with mass of newly created track
        self.active_mass[self._links[:, 0]] -= (1.0 - 0.8) * (
            self.active_mass[self._links[:, 0]] - detections.mass[self._links[:, 1]]
        )
        self.active_mass = torch.cat((self.active_mass[active_mask], detections.mass[self._unmatched_detections]))

        assert len(self.all_positions[-1]) == len(
            self.active_tracks
        ), "Wrong implementation of post_association. It should register the positions of each active tracks"
        for i, track in enumerate(self.active_tracks):
            track.register_track_id(i)

        # Remove the computed features if save_all is False
        if not self.save_all and remove_feats:
            detections.data.pop("features")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable

import numpy as np
import numba  # type: ignore
import torch

import byotrack  # pylint: disable=cyclic-import


class FeaturesExtractor(ABC):
    """Abstract class for features extraction from an image for each detections"""

    @abstractmethod
    def __call__(self, frame: np.ndarray, detections: byotrack.Detections) -> torch.Tensor:
        """Extract the features for each detection given from the current frame

        Args:
            frame (np.ndarray): The video frame
                Shape: ([D, ]H, W, C)
            detections (byotrack.Detections): Detected objects on the frame, for which features should be computed

        Returns:
            torch.Tensor: d features for each detection.
                Shape: (n_dets, d), dtype: float32

        """

    def register(self, frame: np.ndarray, detections: byotrack.Detections) -> None:
        """Register the features inside the detections data.

        It is up to the caller to unregister the features if needed. (It may increase memory usage)

        Args:
            frame (np.ndarray): The video frame
                Shape: ([D, ]H, W, C)
            detections (byotrack.Detections): Detected objects on the frame, for which features should be computed

        """
        detections.data["features"] = self(frame, detections)


class MultiFeaturesExtractor(FeaturesExtractor):
    """Merge multiple features extractors by concatenating the features

    It is useful to test several features together quickly. But this will probably lead
    to sub-optimal computational performances when computing the features.
    """

    def __init__(self, features_extractors: Iterable[FeaturesExtractor]) -> None:
        self.features_extractors = features_extractors

    def __call__(self, frame: np.ndarray, detections: byotrack.Detections) -> torch.Tensor:
        features = [extractor(frame, detections) for extractor in self.features_extractors]
        return torch.cat(features, dim=-1)


class MassExtractor(FeaturesExtractor):
    """Extract the mass of each detection (number of pixels)"""

    def __call__(self, frame: np.ndarray, detections: byotrack.Detections):
        return detections.mass


class IntensityExtractor(FeaturesExtractor):
    """Extract the sum of intensities of each detection"""

    def __call__(self, frame: np.ndarray, detections: byotrack.Detections):
        torch.tensor(compute_intensity(detections.segmentation.numpy(), frame.sum(axis=-1)), dtype=torch.float32)


@numba.njit
def compute_intensity(segmentation: np.ndarray, frame: np.ndarray) -> np.ndarray:
    """Extract the cumulated intensity of each detection

    Args:
        segmentation (np.ndarray): Segmentation mask
        frame (np.ndarray): Video frame (should have the same number of pixels than segmentation)

    Returns:
        np.ndarray: Sum of intensity for each object

    """
    n = segmentation.max()
    intensity = np.zeros(n, dtype=frame.dtype)

    # Ravel in 1D
    segmentation = segmentation.reshape(-1)
    frame = frame.reshape(-1)

    for i in range(segmentation.shape[0]):
        instance = segmentation[i] - 1
        if instance != -1:
            intensity[instance] += frame[i]

    return intensity



"""IOs for CTC datasets [10]

We provide loading and saving functions for detections and tracks.
"""

import os
import pathlib
from typing import Collection, Dict, List, Optional, Union, Sequence, Tuple
import warnings

import numba  # type: ignore
import numpy as np
import tifffile  # type: ignore
import torch
import tqdm.auto as tqdm

import byotrack
import byotrack.utils
import byotrack.video.reader
from byotrack.api.detector.detections import _fast_unique, _position_from_segmentation, relabel_consecutive


class GroundTruthDetector(byotrack.BatchDetector):
    """Converts a 'ground-truth' video of segmentation into byotrack Detections

    The video should not be normalized and each pixel is expected to be an integer.

    Example:

    .. code-block:: python

        # Load the segmentation video, it can be directly loaded from any folder containing tiff images
        video = byotrack.Video("dataset/01_ERR_SEG")  # Load segmentation for CLB
        # video = byotrack.Video("dataset/01_GT/SEG")  # Load ground-truth segmentation
        # video = byotrack.Video("dataset/01_RES/TRA")  # Load predicted tracks segmentation

        detector = GroundTruthDetector()

        detections_sequence = detector.run(video)

    """

    progress_bar_description = "Detections (Load from CTC format)"

    def detect(self, batch: np.ndarray) -> List[byotrack.Detections]:
        assert batch.shape[-1] == 1, "Multichannel segmentation are not supported"
        assert np.issubdtype(batch.dtype, np.integer)

        detections_sequence = []
        for frame in batch:
            detections_sequence.append(
                byotrack.Detections({"segmentation": torch.tensor(frame[..., 0], dtype=torch.int32)})
            )

        return detections_sequence


def _parse_meta_data(file: pathlib.Path) -> Dict[int, Tuple[int, int, int]]:
    """Parse the CTC meta_data (res_track.txt or man_track.txt) into a Dict"""
    meta: Dict[int, Tuple[int, int, int]] = {}

    for line in file.read_text(encoding="utf-8").strip().split("\n"):
        identifier, start, end, parent = (int(word) for word in line.strip().split())
        meta[identifier - 1] = (start, end, parent)

    return meta


def load_tracks(  # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    path: Union[str, os.PathLike],
) -> List[byotrack.Track]:
    """Load saved tracks at the CTC format [10]

    The CTC format for tracks consists of one tiff file for each frame which contains the segmentation
    of active tracks on the frame and a text file containing track ids, start and end frames and parent track.

    First the code parses the segmentation tiff files (either "man*{frame_id}.tif" or "mask{frame_id}.tif")
    and recovers all the known positions (plus the associated detections_ids) of the tracks. Then it parses
    the metadata in the txt file (either "man_track.txt" or "res_track.txt") and validate the tracks creation.

    See the official documentation of CTC at
    https://public.celltrackingchallenge.net/documents/Naming%20and%20file%20content%20conventions.pdf

    Tracks with parent are not supported yet.

    Example:

    .. code-block:: python

        import byotrack
        import byotrack.visualize
        from byotrack.dataset import ctc

        # Load the video and normalize it
        video = byotrack.Video("dataset/01")  # Load videos
        video.set_transform(byotrack.VideoTransformConfig(aggregate=True, normalize=True))

        # Optionnally, load ground-truth segmentations (may take a lot of RAM)
        detections_sequence = ctc.GroundTruthDetector().run(byotrack.Video("dataset/01_GT/TRA"))

        # Load ground-truth tracks
        tracks = ctc.load_tracks("dataset/01_GT/TRA")

        # Visualize everything
        byotrack.visualize.InteractiveVisualizer(video, detections_sequence, tracks)


    Args:
        path (Union[str, os.PathLike]): Path to the tracks data

    Returns:
        List[byotrack.Track]: Saved tracks

    """

    path = pathlib.Path(path)
    is_res = False
    tracks_meta: Dict[int, Tuple[int, int, int]] = {}

    if (path / "res_track.txt").exists():
        is_res = True
        tracks_meta = _parse_meta_data(path / "res_track.txt")
    elif (path / "man_track.txt").exists():
        tracks_meta = _parse_meta_data(path / "man_track.txt")
    else:
        # We can load without it, but not resolve the dividing of tracks
        warnings.warn("res_track.txt or man_track.txt not found. Loading will be done without it.")

    # Dict containing for each track_id, the list of frames found for the track, the associated det_id and its position
    tracks_data: Dict[int, Tuple[List[int], List[int], List[np.ndarray]]] = {}

    # Let's iterate through any found segmentation
    segmentation_paths = path.glob("mask*.tif") if is_res else path.glob("man_*.tif")
    loader = byotrack.video.reader.FrameTiffLoader()

    for path_ in tqdm.tqdm(byotrack.utils.sorted_alphanumeric(segmentation_paths), desc="Loading CTC tracks"):
        if is_res:
            frame_id = int(path_.stem[len("mask") :])
        elif "seg" in path.stem:
            frame_id = int(path_.stem[len("man_seg") :])
        else:
            frame_id = int(path_.stem[len("man_track") :])

        frame = loader(path_)

        assert frame.shape[-1] == 1, "Multichannel segmentation are not supported"
        frame = frame[..., 0]

        # Compute the mapping done by relabel consecutive
        unique = _fast_unique(frame)[1:] - 1  # Remove the background (0)
        positions = _position_from_segmentation(relabel_consecutive(frame, inplace=True))

        for det_id, (track_id, position) in enumerate(zip(unique, positions)):
            if track_id not in tracks_data:
                tracks_data[track_id] = ([], [], [])

            tracks_data[track_id][0].append(frame_id)
            tracks_data[track_id][1].append(det_id)
            tracks_data[track_id][2].append(position)

    # Build tracks
    tracks = []
    for track_id, data in tracks_data.items():
        if track_id not in tracks_meta and tracks_meta:
            warnings.warn(f"Missing identifier {track_id} in the txt metadata.")

        if track_id in tracks_meta:
            start = tracks_meta[track_id][0]
            last = tracks_meta[track_id][1]
            parent = tracks_meta[track_id][2] - 1  # Offset of 1 in identifiers

            if min(data[0]) < start:
                raise ValueError(f"Found track {track_id} on frame {min(data[0])} before for it started")

            if max(data[0]) > last:
                raise ValueError(f"Found track {track_id} on frame {max(data[0])} after for it ended")
        else:
            start, last = min(data[0]), max(data[0])
            parent = -1

        known_frames = torch.tensor(data[0]) - start
        known_detection_ids = torch.tensor(data[1], dtype=torch.int32)
        known_positions = torch.tensor(np.array(data[2]), dtype=torch.float32)

        points = torch.full((last - start + 1, known_positions.shape[1]), torch.nan)
        detection_ids = torch.full((last - start + 1,), -1, dtype=torch.int32)

        detection_ids[known_frames] = known_detection_ids
        points[known_frames] = known_positions

        tracks.append(byotrack.Track(start, points, track_id, detection_ids, parent_id=parent))

    byotrack.Track.check_tracks(tracks, warn=True)

    return tracks


def save_detections(
    path: Union[str, os.PathLike],
    detections_sequence: Sequence[byotrack.Detections],
    as_res=True,
    as_seg=False,
    n_digit=4,
) -> None:
    """Save detections in the CTC format [10]

    It will save one tiff image for each frame containing the segmentation of objects.

    See the official documentation of CTC at
    https://public.celltrackingchallenge.net/documents/Naming%20and%20file%20content%20conventions.pdf

    Args:
        path (Union[str, os.PathLike]): Folder path where to store the .tif files
        detections_sequence (Sequence[byotrack.Detections]): Detections for each frame
        as_res (bool): Whether to store as a results or as a ground-truth.
            Ground-truth are stored as "man_trackT.tif"
            Results as "maskT.tif"
            Default: True
        as_seg (bool): Only for ground-truth: file names are "man_segT.tif" instead of "man_trackT.tif"
            Default: False
        n_digit (int): Number of digit used to encode time in file names.
            Default: 4

    """

    path = pathlib.Path(path)

    os.makedirs(path, exist_ok=True)

    for frame_id, detections in enumerate(tqdm.tqdm(detections_sequence, desc="Saving Detections to CTC")):
        segmentation = detections.segmentation.numpy().astype(np.uint16)

        if as_res:
            name = f"mask{frame_id:0{n_digit}}.tif"
        elif as_seg:
            name = f"man_seg{frame_id:0{n_digit}}.tif"
        else:
            name = f"man_track{frame_id:0{n_digit}}.tif"

        if detections.dim == 2:  # Probably not required here, but let's be precise just in case
            segmentation = segmentation[None, None, None, ..., None]  # ImageJ tiff format: TZCYXS
        else:
            segmentation = segmentation[None, :, None, ..., None]  # ImageJ tiff format: TZCYXS

        tifffile.imwrite(path / name, segmentation, imagej=True)


def _save_metadata(path: pathlib.Path, tracks: Collection[byotrack.Track]):
    """Saves tracks metadata in the given file"""
    lines = []
    for track in tracks:
        if track.merge_id != -1:
            warnings.warn("CTC format do not support merge events. Merge events are removed.")

        lines.append(
            " ".join(
                [
                    str(track.identifier + 1),  # Offset of 1 for track id as 0 is not valid for CTC
                    str(track.start),
                    str(track.start + len(track) - 1),
                    str(track.parent_id + 1),  # Offset of 1 for identifiers
                ]
            )
        )

    path.write_text("\n".join(lines))


@numba.njit(parallel=True)
def _fast_relabel(segmentation: np.ndarray, mapping: np.ndarray):
    """Inplace fast relabel with given mapping

    It assumes that seg.max() is small before the number of pixels of the image (which is always the case in practice).
    """
    segmentation = segmentation.reshape(-1)

    for i in numba.prange(segmentation.size):  # pylint: disable=not-an-iterable
        if segmentation[i]:
            segmentation[i] = mapping[segmentation[i] - 1]


@numba.njit(parallel=True)
def _fast_disk_2d(
    segmentation: np.ndarray,
    bbox: np.ndarray,
    positions: np.ndarray,
    identifiers: np.ndarray,
    radius: np.ndarray,
    *,
    overwrite=False,
):
    """Fast inplace drawing of disk in 2D

    Args:
        segmentation (np.ndarray): Segmentation image to draw on
            Shape: (H, W), dtype: uint16
        bbox (np.ndarray): Indices to consider around the particles (see `draw_disk`)
            Shape: (m, d), dtype: int
        positions (np.ndarray): Positions of the disks' centers
            Shape: (n, d), dtype: float32
        identifiers (np.ndarray): Identifier of each disk
            Shape: (n,), dtype: uint16
        radius (np.ndarray): Radius of each disk
            Shape: (n, ), dtype: float32
        overwrite (bool): Overwrite pixels that are already written (!=0)
            Default: False

    """
    positions_round = np.round(positions).astype(np.int64)
    radius = radius**2
    best_dist = np.full(segmentation.shape, np.inf, dtype=np.float32)

    for k in range(positions_round.shape[0]):
        for l in numba.prange(bbox.shape[0]):  # pylint: disable=not-an-iterable
            pos = bbox[l] + positions_round[k]

            i, j = pos

            if not 0 <= i < segmentation.shape[0] or not 0 <= j < segmentation.shape[1]:
                continue

            if not overwrite and segmentation[i, j] != 0 and best_dist[i, j] == np.inf:
                continue

            delta = pos - positions[k]
            dist = delta @ delta

            if dist <= radius[k]:
                if dist < best_dist[i, j]:
                    segmentation[i, j] = identifiers[k]
                    best_dist[i, j] = dist


@numba.njit(parallel=True)
def _fast_disk_3d(  # pylint: disable=too-many-locals
    segmentation: np.ndarray,
    bbox: np.ndarray,
    positions: np.ndarray,
    identifiers: np.ndarray,
    radius: np.ndarray,
    *,
    anisoptropy=1.0,
    overwrite=False,
):
    """Fast inplace drawing of disk in 3D

    Args:
        segmentation (np.ndarray): Segmentation image to draw on
            Shape: (D, H, W), dtype: uint16
        bbox (np.ndarray): Indices to consider around the particles (see `draw_disk`)
            Shape: (m, d), dtype: int
        positions (np.ndarray): Positions of the disks' centers
            Shape: (n, d), dtype: float32
        identifiers (np.ndarray): Identifier of each disk
            Shape: (n,), dtype: uint16
        radius (np.ndarray): Radius of each disk
            Shape: (n, ), dtype: float32
        anisotropy (float): Relative size of a pixel along the depth dimension
            versus height/width dimensions.
            Default: 1.0
        overwrite (bool): Overwrite pixels that are already written (!=0)
            Default: False

    """
    positions_round = np.round(positions).astype(np.int64)
    radius = radius**2
    best_dist = np.full(segmentation.shape, np.inf, dtype=np.float32)

    for k in range(positions_round.shape[0]):
        for l in numba.prange(bbox.shape[0]):  # pylint: disable=not-an-iterable
            pos = bbox[l] + positions_round[k]

            z, i, j = pos

            if (
                not 0 <= z < segmentation.shape[0]
                or not 0 <= i < segmentation.shape[1]
                or not 0 <= j < segmentation.shape[2]
            ):
                continue

            if not overwrite and segmentation[z, i, j] != 0 and best_dist[z, i, j] == np.inf:
                continue

            delta = pos - positions[k]
            delta[0] *= anisoptropy  # Increase distance in Z by the anisotropy
            dist = delta @ delta

            if dist <= radius[k]:
                if dist < best_dist[z, i, j]:
                    segmentation[z, i, j] = identifiers[k]
                    best_dist[z, i, j] = dist


def draw_disk(
    segmentation: np.ndarray,
    positions: np.ndarray,
    identifiers: np.ndarray,
    radius: np.ndarray,
    *,
    anisotropy=1.0,
    overwrite=False,
):
    """Draw disks on the segmentation

    Args
        segmentation (np.ndarray): Segmentation image to draw on
            Shape: (D, H, W), dtype: uint16
        positions (np.ndarray): Positions of the disks' centers
            Shape: (n, d), dtype: float32
        identifiers (np.ndarray): Identifier of each disk
            Shape: (n,), dtype: uint16
        radius (np.ndarray): Radius of each disk
            Shape: (n, ), dtype: float32)
        anisotropy (float): Relative size of a pixel along the depth dimension
            versus height/width dimensions.
            Default: 1.0
        overwrite (bool): Overwrite pixels that are already written (!=0)
            Default: False

    """
    # Wrapped to redirect in 2D/3D and numba does not support np.indices in 3.8
    if segmentation.ndim == 3:
        thresh = round(radius.max())
        bbox: np.ndarray = np.indices((thresh * 2 + 1, thresh * 2 + 1, thresh * 2 + 1)).transpose(1, 2, 3, 0) - thresh
        bbox = bbox.reshape(-1, 3)
        _fast_disk_3d(segmentation, bbox, positions, identifiers, radius, anisoptropy=anisotropy, overwrite=overwrite)
    else:
        thresh = int(round(radius.max()))
        bbox = np.indices((thresh * 2 + 1, thresh * 2 + 1)).transpose(1, 2, 0) - thresh
        bbox = bbox.reshape(-1, 2)
        _fast_disk_2d(segmentation, bbox, positions, identifiers, radius, overwrite=overwrite)


def save_tracks(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements,too-many-arguments
    path: Union[str, os.PathLike],
    tracks: Collection[byotrack.Track],
    detections_sequence: Sequence[byotrack.Detections] = (),
    *,
    as_res=True,
    as_seg=False,
    default_radius=3.0,
    last=0,
    shape: Optional[Tuple[int, ...]] = None,
    n_digit=4,
    anisotropy=1.0,
    overwrite_detections=False,
):
    """Save tracks in the CTC format [10]

    It will save one tiff image for each frame containing the segmentation of objects and a metadata txt file
    describing the tracks identifiers, start/end frames and parents.

    Parent information is not supported yet.

    When no detections_sequence is given, tracks segmentations are simply drawn as disk with `default_radius`
    at the track localization.
    When detections_sequence is given, then for tracks without detections associated, a disk is drawn
    with `default_radius` (set at 0 to drop this behavior), otherwise the detection segmentation is used.

    For smarter behaviors, one can directly modify the segmentation before saving.

    See the official documentation of CTC at
    https://public.celltrackingchallenge.net/documents/Naming%20and%20file%20content%20conventions.pdf

    Args:
        path (Union[str, os.PathLike]): Folder path where to store the .tif files
        tracks (Collection[byotrack.Track]): Tracks to save
        detections_sequence (Sequence[byotrack.Detections]): Optional detections for each frame
            Default: ()
        as_res (bool): Whether to store as a results or as a ground-truth.
            Ground-truth are stored as "man_trackT.tif" and "man_track.txt"
            Results as "maskT.tif" and "res_track.txt"
            Default: True
        as_seg (bool): Only for ground-truth: file names are "man_segT.tif" instead of "man_trackT.tif"
            Note that it will also store the meta data to allow reloading the tracks.
            Default: False
        default_radius (float): Radius of drawn disk when no segmentation is available.
            Default: 5.0 (pixels)
        last (int): Overwrite last frame to consider
            Default: 0 (Will compute it from the last tracked particles)
        shape (Optional[Tuple[int, ...]]): Optional shape. Required when no detections_sequence is provided
            Default: None
        n_digit (int): Number of digit used to encode time in file names.
            Default: 4
        anisotropy (float): Relative size of a pixel along the depth dimension
            versus height/width dimensions.
            Default: 1.0
        overwrite_detections (bool): Overwrite the segmentation of objects with disk.
            Default: False (Disk are only drawn on background)

    """
    path = pathlib.Path(path)

    os.makedirs(path, exist_ok=True)

    if as_res:
        _save_metadata(path / "res_track.txt", tracks)
    else:
        _save_metadata(path / "man_track.txt", tracks)

    if not last:
        last = max(track.start + len(track) - 1 for track in tracks)

    if detections_sequence:
        if shape:
            assert detections_sequence[0].shape == shape, "Given shape is not compatible with detections one"
        shape = detections_sequence[0].shape
    else:
        assert shape is not None, "Without detections_sequence, you need to provide the shape argument"

    for frame_id in tqdm.trange(last + 1, desc="Saving tracks to CTC"):
        has_detections = len(detections_sequence) > frame_id

        disk_positions = []
        disk_ids = []

        if has_detections:
            det_to_track_ids = np.zeros(detections_sequence[frame_id].length, dtype=np.uint16) - 1

        for track in tracks:
            position = track[frame_id]

            if torch.isnan(position).any():  # Track not defined
                if track.start <= frame_id < track.start + len(track):
                    warnings.warn(
                        "Found a missing position inside a track segment. This is not supported by CTC softwares."
                        "Consider filling holes with Interpolators."
                    )
                continue

            if has_detections:
                detection_id = track.detection_ids[frame_id - track.start]
                if detection_id != -1:
                    det_to_track_ids[detection_id] = track.identifier
                    continue

            disk_positions.append(position)
            disk_ids.append(track.identifier)

        if has_detections:
            segmentation = detections_sequence[frame_id].segmentation.clone().numpy().astype(np.uint16)
            _fast_relabel(segmentation, det_to_track_ids + 1)  # Offset of 1 for track id as 0 is not valid for CTC
        else:
            segmentation = np.zeros(shape, dtype=np.uint16)

        # Add circle to the segmentation
        if disk_ids:
            unique = set(_fast_unique(segmentation).tolist())
            draw_disk(
                segmentation,
                torch.stack(disk_positions).numpy(),
                np.array(disk_ids, dtype=np.uint16) + 1,
                np.full(len(disk_ids), default_radius, dtype=np.float32),
                anisotropy=anisotropy,
                overwrite=overwrite_detections,
            )

            # Safety checks because CTC is quite restrictive
            new_unique = set(_fast_unique(segmentation).tolist())
            diff = {identifier - 1 for identifier in unique.difference(new_unique)}
            if diff:
                warnings.warn(
                    f"Some tracks were fully occluded: {diff}. This will induce"
                    " a missing position in the track segment which is not suppoted by CTC softwares."
                    " Consider decreasing the radius of track disks, or adding manually their segmentations."
                )

            diff = {identifier for identifier in disk_ids if identifier + 1 not in new_unique}
            if diff:
                warnings.warn(
                    f"The disk of some added tracks are not found (outside of image or occluded): {diff}."
                    " This will induce a missing position in the track segment which is not suppoted by CTC softwares."
                    " Fixing this can be done by changing the radius, or manually dropping/correcting the track."
                )

        # Save a tiff
        if len(shape) == 2:
            segmentation = segmentation[None, None, None, ..., None]  # ImageJ tiff format: TZCYXS
        else:
            segmentation = segmentation[None, :, None, ..., None]  # ImageJ tiff format: TZCYXS

        if as_res:
            name = f"mask{frame_id:0{n_digit}}.tif"
        elif as_seg:
            name = f"man_seg{frame_id:0{n_digit}}.tif"
        else:
            name = f"man_track{frame_id:0{n_digit}}.tif"

        tifffile.imwrite(path / name, segmentation, imagej=True)



"""Wrappers around CTC metrics [10]

We currently only wrap the CTC software providing SEG, DET and TRA metrics.

In the future, we may wrap the Fiji plugin to access BIO metrics.
"""

import os
import pathlib
import platform
import re
import shutil
import subprocess
import tempfile
from typing import Collection, Union, Sequence
import warnings


import byotrack
from byotrack import fiji
from byotrack.dataset import ctc
from byotrack.utils import sorted_alphanumeric


class CTCSoftwareRunner:  # pylint: disable=too-few-public-methods
    """Runs CTC software (TRA/DET/SEG) [10]

    It wraps the TRA/DET/SEG softwares that are distributed at https://celltrackingchallenge.net/evaluation-methodology
    It is up to the user to download the correct executable and ensure that it works.

    These softwares take as inputs the path to the folder of the dataset (with ground-truth and results tracks),
    the sequence id to evaluate and the number of digits used to encode time in file names. (3 or 4)

    They output a string with the measure score, that we parse.

    Note:
        This code requires CTC softwares as extra dependencies.
        They can be downloaded at https://celltrackingchallenge.net/evaluation-methodology.

    Attributes:
        ctc_software (pathlib.Path): Path to the ctc software folder.
            It should be the root folder containing Win/Linux/Mac subfolders with their executables.
        system (str): The user system in the CTC nomenclature. One of Linux, Mac or Win.
        last_log (str): Logs of the last computed metrics

    """

    cmd = './{executable} "{dataset}" {seq:02} {n_digit}'

    def __init__(self, ctc_software: Union[str, os.PathLike]):
        self.ctc_software = pathlib.Path(ctc_software)
        self.last_log = ""

        self.system = platform.system()
        if self.system == "Darwin":
            self.system = "Mac"
        elif self.system == "Windows":
            self.system = "Win"
        # else: Linux is already good, Java is not supported so let's just try Linux ?

        assert self.ctc_software.is_dir(), "The path to the unzipped CTC `evaluation_software` is required"
        assert (self.ctc_software / self.system).is_dir(), f"Softwares for your system ({self.system}) are not found"

    def run(self, metric: str, dataset: Union[str, os.PathLike], seq=1, n_digit=4) -> float:
        """Run the CTC software for the given metric and dataset

        Args:
            metric (str): The metric to evaluate. One of (TRA, DET, SEG).
            dataset (Union[str, os.PathLike]): Path to the dataset to evaluate.
            seq (int): Sequence to evaluate inside the dataset.
                CTC softwares will compare {dataset}/{seq:02}_RES with {dataset}/{seq:02}_GT
                Default: 1
            n_digit (int): Number of digits used to encode time in file names.
                It is dataset dependant, but in ByoTrack, by default we use 4 digits.
                Default: 4

        Returns:
            float: The evaluated metric

        """
        assert metric in ("SEG", "DET", "TRA"), "Only SEG, DET and TRA metrics are available with this software"

        executable = metric + "Measure"
        if self.system == "WIN":
            executable = executable + ".exe"

        cmd = self.cmd.format(executable=executable, dataset=dataset, seq=seq, n_digit=n_digit)

        if self.system == "WIN":
            cmd = cmd[2:]  # Strip ./ on windows

        print("Calling CTC softwares with:", cmd)
        outputs = subprocess.run(
            cmd, check=False, cwd=f"{self.ctc_software}/{self.system}", shell=True, stdout=subprocess.PIPE
        )
        output_string = outputs.stdout.decode("utf-8")

        # Let's parse the output
        try:
            value = float(output_string.split()[-1])
        except Exception as exc:  # pylint: disable=broad-except
            raise RuntimeError(
                "Cannot parse outputs, the CTC software probably found an error: " + output_string
            ) from exc

        self.last_log = (pathlib.Path(dataset) / f"{seq:02}_RES" / f"{metric}_log.txt").read_text("utf-8")

        return value


class CTCMetrics(CTCSoftwareRunner):
    """Wrapper around CTC metrics. [10]

    It wraps the TRA/DET/SEG softwares that are distributed at https://celltrackingchallenge.net/evaluation-methodology
    It is up to the user to download the correct executable and ensure that it works.

    To compute a metric, the code creates a temporary folder, where it saves the ground-truth tracks
    alongside the predicted tracks. Then it runs the required CTC software, and finally remove the
    temporary folder. See `CTCSoftwareRunner`.

    Note:
        This code requires CTC softwares as extra dependencies.
        They can be downloaded at https://celltrackingchallenge.net/evaluation-methodology.

    We provide two distincts methods to evaluate the metrics. One based on tracks, the other on detections:

    1. `compute_detection_metrics` compute the DET and SEG using the results of a segmentation algorithm (no tracks
    information is required). Note that TRA cannot be computed in this case.
    2. `compute_tracking_metrics` compute the metrics for tracks. One can compute the DET and SEG for the tracks,
    but these will not match with the ones computed for detections. Indeed, tracks filter out some detections
    and may add points to tracks to fill missing detections. Usually, this leads to better DET/SEG metrics
    as it reduces false positive and false negative.

    It will store the logs of the last called metric in `self.last_log`

    """

    def compute_tracking_metric(
        self,
        metric: str,
        tracks: Collection[byotrack.Track],
        ground_truth_tracks: Union[str, os.PathLike, Collection[byotrack.Track]],
        *,
        detections_sequence: Sequence[byotrack.Detections] = (),
        ground_truth_detections_sequence: Sequence[byotrack.Detections] = (),
        **kwargs,
    ) -> float:
        """Compute the given metric for the given tracks and ground truthes.

        It will create a temporary folder, to store the tracks and ground-truthes in the right format and then
        execute the CTC software. The temporary folder is removed at the end.

        See `run` to simply compute the metric of an already existing folder.

        Note:
            In CTC, matching of predicted tracks with GT ones is done based on a kind of IOU.
            Therefore, it may be useful to provide the detections_sequence associated with the tracks.
            You may also tweak the `default_radius` arguments (in kwargs). (See `ctc.save_tracks`)

        Note:
            This function computes the tracking metrics, which are different than
            those computed with `compute_detection_metric` for SEG and DET.
            For tracking metrics, we export the segmentation of tracks, rather than the detections_sequence.
            Tracks segmentation differs from the predicted segmentation, as the tracks processus
            filters out detections and may predict tracks position for missing ones. (It usually improves
            the segmentation metrics, as it reduces false positive and false negative)

        Args:
            metric (str): The metric to evaluate. One of (TRA, DET, SEG).
            tracks (Collection[byotrack.Track]): Predicted tracks to evaluate.
            ground_truth_tracks (Union[str, os.PathLike, Collection[byotrack.Track]]): Ground truth data.
                It is either a path to the GT tracks folder, which will be copied in our temporary folder.
                Or it is a list of ByoTrack.Track, that will be saved in the temporary folder.
            detections_sequence (Sequence[byotrack.Detections]): Optional detections, used when saving the tracks.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.
            ground_truth_detections_sequence (Sequence[byotrack.Detections]): Optional detections for ground-truth.
                When saving the GT tracks, these detections are used. See `ctc.save_tracks`.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.
            **kwargs: Additional arguments to provide to the `ctc.save_tracks` function.
                'shape': Provide the shape of saved image. It is mandatory if no detections is provided.
                'default_radius': Radius of the disk drawn for tracks that have no detections.
                'last': Last frame to consider. (Typically to shorten the sequences,
                or if no object is tracked on the last frames, this will enforce the creation of empty tiff files)

        """
        if "shape" in kwargs:
            shape = kwargs.pop("shape")
        else:
            if detections_sequence:
                shape = detections_sequence[0].shape
            elif ground_truth_detections_sequence:
                shape = ground_truth_detections_sequence[0].shape
            else:
                raise ValueError("Without any detections, `shape` must be provided")

        with tempfile.TemporaryDirectory(prefix="ByoTrack-CTC-Metrics") as output_dir:
            output_path = pathlib.Path(output_dir)
            ground_truth_path = output_path / "01_GT" / ("SEG" if metric == "SEG" else "TRA")
            ctc.save_tracks(
                output_path / "01_RES", tracks, detections_sequence, as_res=True, shape=shape, n_digit=4, **kwargs
            )
            if isinstance(ground_truth_tracks, (str, os.PathLike)):
                if ground_truth_detections_sequence:
                    warnings.warn(
                        "When using a saved GT folder, it will be copied and ground-truth detections are ignored"
                    )
                self.copy_ground_truth(ground_truth_tracks, ground_truth_path, as_seg=metric == "SEG")
            else:
                ctc.save_tracks(
                    ground_truth_path,
                    ground_truth_tracks,
                    ground_truth_detections_sequence,
                    as_res=False,
                    as_seg=metric == "SEG",
                    shape=shape,
                    **kwargs,
                )

            return self.run(metric, output_path, 1)

    def compute_detection_metric(
        self,
        metric: str,
        detections_sequence: Sequence[byotrack.Detections],
        ground_truth_detections_sequence: Union[str, os.PathLike, Sequence[byotrack.Detections]],
    ) -> float:
        """Compute the given metric for the given detections and ground truthes.

        It will create a temporary folder, to store the detections and ground-truthes in the right format and then
        execute the CTC software. The temporary folder is removed at the end.

        See `run` to simply compute the metric of an already existing folder.

        Note:
            This function truly computes the detections metrics, which are different than
            those computed with `compute_tracking_metric` for SEG and DET.
            For tracking metrics, we export the segmentation of tracks, rather than the detections_sequence.
            Tracks segmentation differs from the predicted segmentation, as the tracks processus
            filters out detections and may predict tracks position for missing ones. (It usually improves
            the segmentation metrics, as it reduces false positive and false negative)

        Args:
            metric (str): The metric to evaluate. One of (TRA, DET, SEG).
            tracks (Collection[byotrack.Track]): Predicted tracks to evaluate.
            ground_truth_tracks (Union[str, os.PathLike, Collection[byotrack.Track]]): Ground truth data.
                It is either a path to the GT tracks folder (), which will be copied in our temporary folder.
                Or it is a list of ByoTrack.Track, that will be saved in the temporary folder.
            detections_sequence (Sequence[byotrack.Detections]): Optional detections, used when saving the tracks.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.
            ground_truth_detections_sequence (Sequence[byotrack.Detections]): Optional detections for ground-truth.
                When saving the GT tracks, these detections are used. See `ctc.save_tracks`.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.

        """
        assert metric != "TRA", "TRA metric cannot be computed for detections"

        with tempfile.TemporaryDirectory(prefix="ByoTrack-CTC-Metrics") as output_dir:
            output_path = pathlib.Path(output_dir)
            ground_truth_path = output_path / "01_GT" / ("SEG" if metric == "SEG" else "TRA")
            ctc.save_detections(output_path / "01_RES", detections_sequence)
            if isinstance(ground_truth_detections_sequence, (str, os.PathLike)):
                self.copy_ground_truth(ground_truth_detections_sequence, ground_truth_path, as_seg=metric == "SEG")
            else:
                ctc.save_detections(
                    ground_truth_path, ground_truth_detections_sequence, as_res=False, as_seg=metric == "SEG"
                )

            return self.run(metric, output_path, 1)

    @staticmethod
    def copy_ground_truth(
        ground_truth_path: Union[str, os.PathLike], target_path: Union[str, os.PathLike], as_seg=False
    ) -> None:
        """Copy the ground-truth to the target

        It will copy any files matching maskT.tif, res_track.txt, man_trackT.tif, man_segT.tif, man_track.txt.
        Nomenclature of files is adapted to match ground truth one. Also, files with 3 digits
        to encode the time information, will be saved with 4 digits.

        Note:
            It is probably more permissive than the software, allowing to convert from/to SEG to/from TRA annotations.
            It also supports converting results into annotations (SEG and TRA).
            In some cases, it may produce invalid data for CTC softwares, for instance converting SEG GT into TRA ones
            only works if SEG GT are fully annotated and it only allows to run DET measures, as the
            metadata txt file would be missing.

        Args:
            ground_truth_path (Union[str, os.PathLike]): Path to ground truthes.
                We support any directory containing potential files to copy. (In particular
                it can be 01_GT/SEG or 01_GT/TRA)
            target_path (Union[str, os.PathLike]): Path where to copy files.
            as_seg (bool): Save target_files as man_segT.tif instead of man_trackT.tif
                Default: False
        """

        ground_truth_path = pathlib.Path(ground_truth_path)
        target_path = pathlib.Path(target_path)
        target_path.mkdir(parents=True, exist_ok=True)

        # Search for tiff files
        gt_seg_tiff_paths = sorted_alphanumeric(ground_truth_path.glob("man_seg*.tif"))
        gt_track_tiff_paths = sorted_alphanumeric(ground_truth_path.glob("man_track*.tif"))
        res_tiff_paths = sorted_alphanumeric(ground_truth_path.glob("mask*.tif"))

        # Find the inputs type: either res or annotations, and either SEG or TRA for annotations
        if len(gt_seg_tiff_paths) and len(gt_track_tiff_paths):
            warnings.warn(
                "Found both man_trackT.tif and man_segT.tif files in the GT folder. The most common ones will be copied"
            )

        is_seg = len(gt_seg_tiff_paths) > len(gt_track_tiff_paths)

        if max(len(gt_seg_tiff_paths), len(gt_track_tiff_paths)) and len(res_tiff_paths):
            warnings.warn(
                "Found both ground truth and results tiff files in the GT folder. The most common ones will be used."
            )

        is_res = len(res_tiff_paths) > max(len(gt_seg_tiff_paths), len(gt_track_tiff_paths))

        # Handle optional metadata txt file required for TRA
        if is_res:
            if (ground_truth_path / "res_track.txt").exists():
                shutil.copy(ground_truth_path / "res_track.txt", target_path / "man_track.txt")
        else:
            if (ground_truth_path / "man_track.txt").exists():
                shutil.copy(ground_truth_path / "man_track.txt", target_path / "man_track.txt")

        # Handle SEG case on its own
        if is_seg and not is_res:
            for path in gt_seg_tiff_paths:
                # Parse file name, as it can be man_segT.tif or sometimes man_seg_T_Z.tif in 3D
                stem = path.stem[len("man_seg") :]

                if "_" in stem:
                    assert as_seg, "Cannot convert a partial SEG ground truth into a valid DET/TRA ones"
                    frame_id, slice_id = (int(word) for word in stem[1:].split("_"))
                    shutil.copy(path, target_path / f"man_seg_{frame_id:04}_{slice_id:04}.tif")
                else:
                    frame_id = int(stem)
                    shutil.copy(path, target_path / f"man_{'seg' if as_seg else 'track'}{frame_id:04}.tif")

            return

        # Finally handle the rest
        for path in res_tiff_paths if is_res else gt_track_tiff_paths:
            frame_id = int(path.stem[4 if is_res else 9 :])
            shutil.copy(path, target_path / f"man_{'seg' if as_seg else 'track'}{frame_id:04}.tif")


class BioMetrics:
    """Wrapper around the CTC "Biological measures" Fiji plugin [10]

    It allows the computations of BIO metrics for the Cell Tracking Challenge [10].

    Note:
        This implementation requires Fiji to be installed (https://imagej.net/downloads)
        with the CTC plugins (https://github.com/CellTrackingChallenge/fiji-plugins)

    Attributes:
        runner (byotrack.fiji.FijiRunner): Fiji runner
        last_metrics (Tuple[float, float, float, float]): Sub metrics information for the last computed BIO
            It consists of ("CT", "TF", "BC(i)", "CCA"). Their average gives the BIO metric.

    """

    plugin_name = "Biological measures"
    # Reg exp for parsing outputs: find the last 4 numbers in the output line such in the format 'a, b, c, d]]'
    output_regexp = re.compile(r"([+-]?\d*\.\d+), ([+-]?\d*\.\d+), ([+-]?\d*\.\d+), ([+-]?\d*\.\d+)]]$")

    def __init__(self, fiji_path: Union[str, os.PathLike]) -> None:
        """Constructor

        Args:
            fiji_path (str | os.PathLike): Path to the fiji executable
                The executable can be found inside the installation folder of Fiji.
                Linux: Fiji.app/ImageJ-<os>
                Windows: Fiji.app/ImageJ-<os>.exe
                MacOs: Fiji.app/Contents/MacOs/ImageJ-<os>

        """
        self.runner = fiji.FijiRunner(fiji_path, capture_outputs=True)
        self.last_metrics = (0.0, 0.0, 0.0, 0.0)

    def run(self, dataset: Union[str, os.PathLike], seq=1, n_digit=4) -> float:
        """Run the CTC "Biological measures" plugin on the given dataset.

        The dataset should already have results stored in it. It expects the CTC format.

        Args:
            metric (str): The metric to evaluate. One of (TRA, DET, SEG).
            dataset (Union[str, os.PathLike]): Path to the dataset to evaluate.
            seq (int): Sequence to evaluate inside the dataset.
                The plugin will compare {dataset}/{seq:02}_RES with {dataset}/{seq:02}_GT
                Default: 1
            n_digit (int): Number of digits used to encode time in file names.
                It is dataset dependant, but in ByoTrack, by default we use 4 digits.
                Default: 4

        Returns:
            float: The evaluated metric

        """
        path = pathlib.Path(dataset)

        # Let's do output redirection
        self.runner.run(
            "Biological measures", resPath=path / f"{seq:02}_RES", gtPath=path / f"{seq:02}_GT", noOfDigits=n_digit
        )

        line = self.runner.last_outputs.stdout.decode().strip().split("\n")[-1]
        match = self.output_regexp.search(line)
        if not match:
            raise RuntimeError("Cannot parse outputs, the CTC software probably found an error: " + line)

        self.last_metrics = tuple(  # type: ignore
            (float(group) if float(group) >= 0.0 else 0.0) for group in match.groups()
        )

        return sum(self.last_metrics) / 4

    def compute(
        self,
        tracks: Collection[byotrack.Track],
        ground_truth_tracks: Union[str, os.PathLike, Collection[byotrack.Track]],
        *,
        detections_sequence: Sequence[byotrack.Detections] = (),
        ground_truth_detections_sequence: Sequence[byotrack.Detections] = (),
        **kwargs,
    ) -> float:
        """Compute BIO metric for the given tracks and ground truthes.

        It will create a temporary folder, to store the tracks and ground-truthes in the right format and then
        execute the CTC plugins. The temporary folder is removed at the end.

        See `run` to simply compute the metric of an already existing folder.

        Note:
            In CTC, matching of predicted tracks with GT ones is done based on a kind of IOU.
            Therefore, it may be useful to provide the detections_sequence associated with the tracks.
            You may also tweak the `default_radius` arguments (in kwargs). (See `ctc.save_tracks`)

        Args:
            tracks (Collection[byotrack.Track]): Predicted tracks to evaluate.
            ground_truth_tracks (Union[str, os.PathLike, Collection[byotrack.Track]]): Ground truth data.
                It is either a path to the GT tracks folder, which will be copied in our temporary folder.
                Or it is a list of ByoTrack.Track, that will be saved in the temporary folder.
            detections_sequence (Sequence[byotrack.Detections]): Optional detections, used when saving the tracks.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.
            ground_truth_detections_sequence (Sequence[byotrack.Detections]): Optional detections for ground-truth.
                When saving the GT tracks, these detections are used. See `ctc.save_tracks`.
                Default: ()  # No detections and tracks segmentations will be disk of radius `default_radius`.
            **kwargs: Additional arguments to provide to the `ctc.save_tracks` function.
                'shape': Provide the shape of saved image. It is mandatory if no detections is provided.
                'default_radius': Radius of the disk drawn for tracks that have no detections.
                'last': Last frame to consider. (Typically to shorten the sequences,
                or if no object is tracked on the last frames, this will enforce the creation of empty tiff files)

        """
        if "shape" in kwargs:
            shape = kwargs.pop("shape")
        else:
            if detections_sequence:
                shape = detections_sequence[0].shape
            elif ground_truth_detections_sequence:
                shape = ground_truth_detections_sequence[0].shape
            else:
                raise ValueError("Without any detections, `shape` must be provided")

        with tempfile.TemporaryDirectory(prefix="ByoTrack-CTC-Metrics") as output_dir:
            output_path = pathlib.Path(output_dir)
            ground_truth_path = output_path / "01_GT" / "TRA"
            ctc.save_tracks(
                output_path / "01_RES", tracks, detections_sequence, as_res=True, shape=shape, n_digit=4, **kwargs
            )
            if isinstance(ground_truth_tracks, (str, os.PathLike)):
                if ground_truth_detections_sequence:
                    warnings.warn(
                        "When using a saved GT folder, it will be copied and ground-truth detections are ignored"
                    )
                CTCMetrics.copy_ground_truth(ground_truth_tracks, ground_truth_path)
            else:
                ctc.save_tracks(
                    ground_truth_path,
                    ground_truth_tracks,
                    ground_truth_detections_sequence,
                    as_res=False,
                    shape=shape,
                    **kwargs,
                )

            return self.run(output_path, 1)
